/*
** etk.c
**
** Implements a subset of tcl to execute erlang
** functions (as scripts)
**
**  Works with: Tk4.2 (& Tcl7.6)
*/
#include <stdio.h>
#include <tk.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "etk.h"
#include "tkstr.h"

#if (TK_MAJOR_VERSION != 4) && (TK_MINOR_VERSION != 2)
#error "bad tk version should be 4.2"
#endif

#if (TCL_MAJOR_VERSION != 7) && (TCL_MINOR_VERSION != 6)
#error "bad tcl version should be 7.6"
#endif

typedef unsigned char byte;

/* #define DEBUG */

#define ERL_INPUT  0
#define ERL_OUTPUT 1

static int erl_closed = 0;

static FILE* logf = NULL;  /* File to write error's to */


#define ERL_PACKET_HEADER 4

#define IO_BUFFER_SIZE 32*1024

byte iobuffer[IO_BUFFER_SIZE];


extern int matherr();
int *tclDummyMathPtr = (int *) matherr;

EXTERN int Tcl_LinkVar();
int (*tclDummyLinkVarPtr)() = Tcl_LinkVar;

static Tk_Window mainWindow;	/* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */
static Tcl_Interp *interp;	/* Interpreter for this application. */
static int synchronize ;
static int withdrawn;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;
static char *colormap;
static char *visual;
static char *logfile = NULL;
static int rest = 0;

static Tk_ArgvInfo argTable[] = {
    {"-colormap", TK_ARGV_STRING, (char *) NULL, (char *) &colormap,
        "Colormap for main window"},
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
        "Display to use"},
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
        "Initial geometry for window"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
        "Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
        "Use synchronous mode for display server"},
    {"-visual", TK_ARGV_STRING, (char *) NULL, (char *) &visual,
        "Visual for main window"},
    /* Added for Etk */
    {"-withdrawn", TK_ARGV_CONSTANT, (char*) 1, (char *) &withdrawn,
	 "Start erlkt main window withdrawn" },
    {"-logfile", TK_ARGV_STRING, (char*) NULL, (char *) &logfile,
	 "Logfile for error output" },
    {"--", TK_ARGV_REST, (char *) 1, (char *) &rest,
        "Pass all remaining arguments through to script"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
        (char *) NULL}
};

/*
** Log Error Messages
*/
void LogError(mesg, cmd, argc, argv)
char* mesg; int cmd; int argc; char** argv;
{
    int i;

    if (logf == NULL)
	return;

    fprintf(logf, "%s:", mesg);
    switch(cmd) {
    case TK_ERL_EVENT:  fprintf(logf, " EVENT:"); break;
    case TK_ERL_INVOKE: fprintf(logf, " INVOKE:"); break;
    case TK_ERL_ERROR:  fprintf(logf, " ERROR:"); break;
    }
    if (argc != 0) {
	for (i = 0; i < argc; i++)
	    fprintf(logf, " %s", argv[i]);
    }
    fprintf(logf, "\n");
}


void CloseLog()
{
    time_t t;

    if (logf == NULL)
	return;
    t = time(0);
    fprintf(logf, "TK STOPPED: %s", ctime(&t));
    fclose(logf);
    logf = NULL;
}

void OpenLog(file)
char* file;
{
    if (logf != NULL)
	CloseLog();
    if ((file != NULL) && (*file != '\0')) {
	time_t t;
	if ((logf = fopen(file, "w")) == NULL) {
	    fprintf(stderr, "etk: warning : could not open logfile %s\n",
		    file);
	    return;
	}
	t = time(0);
	fprintf(logf, "TK STARTED: %s", ctime(&t));
    }
}

/*
**
*/
void ErlTk_Exit(code, msg)
int code; char* msg;
{
    char* argv[1];

    if (code != 0) {
	argv[0] = msg;
	LogError("EXIT: Terminating on ", -1, 1, argv[0]);
	CloseLog();
	Tcl_Exit(code);
    }
    else {
	LogError("EXIT: Terminating normal ", -1, 0, 0);
	CloseLog();
	Tcl_Exit(0);
    }
}


/* erlang function are descibed by fun:id */
/* the function it self is stored in an ets data base */

int ErlTk_Execute(cmd, data, interp, argc, argv)
int cmd; ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    switch(cmd) {
    case ERL_TK_BIND:
	return Tk_BindCmd(data, interp, argc, argv);

    case ERL_TK_DESTROY:
	return Tk_DestroyCmd(data, interp, argc, argv);

    case ERL_TK_LOWER:
	return Tk_LowerCmd(data, interp, argc, argv);
	
    case ERL_TK_RAISE:
	return Tk_RaiseCmd(data, interp, argc, argv);
	
    case ERL_TK_BELL:
	return Tk_BellCmd(data, interp, argc, argv);
	
    case ERL_TK_BUTTON:
	return Tk_ButtonCmd(data, interp, argc, argv);
	
    case ERL_TK_CHECKBUTTON:
	return Tk_CheckbuttonCmd(data, interp, argc, argv);
	
    case ERL_TK_RADIOBUTTON:
	return Tk_RadiobuttonCmd(data, interp, argc, argv);
	
    case ERL_TK_LABEL:
	return Tk_LabelCmd(data, interp, argc, argv);
	
    case ERL_TK_UPDATE:
	return Tk_UpdateCmd(data, interp, argc, argv);
	
    case ERL_TK_WINFO:
	return Tk_WinfoCmd(data, interp, argc, argv);
	
    case ERL_TK_WM:
	return Tk_WmCmd(data, interp, argc, argv);
	
    case ERL_TK_BINDTAGS:
	return Tk_BindtagsCmd(data, interp, argc, argv);
	
    case ERL_TK_CANVAS:
	return Tk_CanvasCmd(data, interp, argc, argv);
	
    case ERL_TK_CLIPBOARD:
	return Tk_ClipboardCmd(data, interp, argc, argv);
	
    case ERL_TK_ENTRY:
	return Tk_EntryCmd(data, interp, argc, argv);
	
    case ERL_TK_FRAME:
	return Tk_FrameCmd(data, interp, argc, argv);

    case ERL_TK_TOPLEVEL:
	return Tk_ToplevelCmd(data, interp, argc, argv);
	
    case ERL_TK_FOCUS:
	return Tk_FocusCmd(data, interp, argc, argv);
	
    case ERL_TK_GRAB:
	return Tk_GrabCmd(data, interp, argc, argv);
	
    case ERL_TK_IMAGE:
	return Tk_ImageCmd(data, interp, argc, argv);
	
    case ERL_TK_LISTBOX:
	return Tk_ListboxCmd(data, interp, argc, argv);
	
    case ERL_TK_MENU:
	return Tk_MenuCmd(data, interp, argc, argv);
	
    case ERL_TK_MENUBUTTON:
	return Tk_MenubuttonCmd(data, interp, argc, argv);
	
    case ERL_TK_MESSAGE:
	return Tk_MessageCmd(data, interp, argc, argv);
	
    case ERL_TK_OPTION:
	return Tk_OptionCmd(data, interp, argc, argv);
	
    case ERL_TK_PACK:
	return Tk_PackCmd(data, interp, argc, argv);
	
    case ERL_TK_PLACE:
	return Tk_PlaceCmd(data, interp, argc, argv);
	
    case ERL_TK_SCALE:
	return Tk_ScaleCmd(data, interp, argc, argv);
	
    case ERL_TK_SCROLLBAR:
	return Tk_ScrollbarCmd(data, interp, argc, argv);
	
    case ERL_TK_SELECTION:
	return Tk_SelectionCmd(data, interp, argc, argv);
	
    case ERL_TK_TEXT:
	return Tk_TextCmd(data, interp, argc, argv);

    case ERL_TK_GRID:
	return Tk_GridCmd(data, interp, argc, argv);
	
    case ERL_TK_TK:
	return Tk_TkCmd(data, interp, argc, argv);
	
    case ERL_TK_CMD: {
	Tcl_CmdInfo info;

	if (Tcl_GetCommandInfo(interp, argv[0], &info))
	    return (*info.proc)(info.clientData, interp, argc, argv);
	else {
	    Tcl_SetResult(interp, "widget not found", TCL_STATIC);
	    return TCL_ERROR;
	}
    }

    case ERL_TK_GETVAR: {
	char *value = Tcl_GetVar(interp, argv[1], TCL_GLOBAL_ONLY);
	if (value == NULL)
	    value = "";
	Tcl_AppendResult(interp, value, (char*) NULL);
	return TCL_OK;
    }

    case ERL_TK_SETVAR:
	if (Tcl_SetVar(interp, argv[1], argv[2],
		       TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) == NULL) {
	    return TCL_ERROR;
	}
	return TCL_OK;

    case ERL_TK_EVENT:
	return Tk_EventCmd(data, interp, argc, argv);

    default:
	ErlTk_Exit(1, "unknown etk command");
	return 0;
    }
}


static int erl_read_fill(int fd, char* buf, int len)
{
  int i, got=0;
  
  do {
    if ((i = read(fd, buf+got, len-got)) <= 0)
      return (i);
    got += i;
  } while (got < len);
  return (len);

}

static int erl_write_fill(int fd, char *buf, int len)
{
  int i,done=0;
  
  do {
    if ((i = write(fd, buf+done, len-done)) <= 0)
      return (i);
    done += i;
  } while (done < len);
  return (len);

}

/*
** Send argv info to erlang as:
**
**  32-bit-length  data
*/

int SendToErlang(cmd, clientData, interp, argc, argv)
int cmd; ClientData clientData; Tcl_Interp* interp; int argc; char** argv;
{
    int i;
    int len;
    int pos;
    byte* ptr = iobuffer;
    byte lenbuf[4];

    if (erl_closed) {
	LogError("ERROR: closed ", cmd,  argc, argv);
	return TCL_OK;
    }

    PUT_INT16(ptr, 0); 
    ptr += 2;
    PUT_INT16(ptr, cmd); 
    ptr += 2;

    for (i = 1; i < argc; i++) {
	int n = strlen(argv[i]);
	PUT_INT32(ptr, n); ptr += 4;
	memcpy(ptr, argv[i], n);
	ptr += n;
    }
    len = (ptr - iobuffer);

    switch(ERL_PACKET_HEADER) {
    case 1:
	PUT_INT8(lenbuf, len);
	break;
    case 2:
	PUT_INT16(lenbuf, len);
	break;
    case 4:
	PUT_INT32(lenbuf, len);
	break;
    }
    erl_write_fill(ERL_OUTPUT, lenbuf, ERL_PACKET_HEADER);
    erl_write_fill(ERL_OUTPUT, iobuffer, len);

    return TCL_OK;
}

/*
** Callback for tkerror & bgerror
*/
int tkError(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    LogError("ERROR:", TK_ERL_TKERROR, argc, argv);
    return SendToErlang(TK_ERL_TKERROR, data, interp, argc, argv);
}

/*
** Callback for tkscreenchanged
*/
int tkScreenChanged(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_TKSCREEN, data, interp, argc, argv);
}

/*
** Callback for event functions
** args:  event Id argv[2] argv[3] ... 
** write  Id, argv[2], argv[3] ...
*/

int ErlangEvent(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_EVENT, data, interp, argc, argv);
}

/*
** Callback for invoke functions
** args:  invoke Id argv[2] argv[3] ...
** write  Id, argv[2], argv[3] ...
*/

int ErlangInvoke(data, interp, argc, argv)
ClientData data; Tcl_Interp* interp; int argc; char** argv;
{
    return SendToErlang(TK_ERL_INVOKE, data, interp, argc, argv);
}

/*
**  INPUT:
**  +-----------------+
**  | Sequence number |   2 bytes
**  +-----------------+
**  | Command         |   2 bytes
**  +-----------------+
**  | Argc = n        |   4 bytes
**  +-----------------+
**  | OP | Argv[0]    |   4 bytes
**  +-----------------+
**  | OP | Argv[1]    |   4 bytes
**  +-----------------+
**  |       ...       |
**  +-----------------+
**  | OP | Argv[n-1]  |   4 bytes
**  +-----------------+
**  |  0   0   0   0  |
**  +-----------------+
**  | Len(4) Data     |
**  | Len(4) Data     |
**  | ...             |
**  |                 |
**  +-----------------+
**
**
**  OUTPUT:
**  +-----------------+
**  | Sequence number |   2 bytes
**  +-----------------+
**  | Reply code      |   2 bytes
**  +-----------------+
**  | Reply string    |
**  |     ...         |
**  +-----------------+
**
*/

void ErlangInputProc(data, mask)
ClientData data; int mask;
{
    int fd = (int) data;
    int len;
    int reslen;
    int n;
    int argc;
    int voffs;
    int cmd;
    int seq;
    int code;
    int i;
    byte* p;
    byte* ptr;
    char** argv;
    byte lenbuf[4];

    if ((n = erl_read_fill(fd, (char*)lenbuf, ERL_PACKET_HEADER)) < 0) {
	erl_closed = 1;
	ErlTk_Exit(1, "read error");
    }
    else if (n == 0) {
	erl_closed = 1;
	ErlTk_Exit(0, "eof");
    }
    switch(ERL_PACKET_HEADER) {
    case 1:
	len = GET_INT8(lenbuf);
	break;
    case 2:
	len = GET_INT16(lenbuf);
	break;
    case 4:
	len = GET_INT32(lenbuf);
	break;
    }

    if ((n = erl_read_fill(fd, (char*) iobuffer, len)) < 0)
	ErlTk_Exit(1, "read error");
    else if (n == 0)
	ErlTk_Exit(0, "eof");

    p = iobuffer;
    seq = GET_INT16(p); p += 2;
    cmd = GET_INT16(p); p += 2;
    argc = GET_INT32(p); p += 4;
    argv = (char**) (p);
    ptr  = p + 4*(argc+1);   /* data area pointer */
    voffs = argc*4;

    for (i = 0; i < argc; i++) {
	int offs;

	p = (byte*) (argv + i);
	switch (*p++) {
	case TK_OP_OFFSET:
	    len = GET_INT32(ptr);  /* get data length */
	    ptr += 4;              /* skip length field */
	    argv[i] = ptr;
	    ptr += len;            /* prepare for next data */
	    break;
	case TK_OP_OPTION:
	    offs = GET_INT24(p);
	    if (offs >= OPT_MAX)
		ErlTk_Exit(1, "bad option offset");
	    argv[i] = tkopt[offs];
	    break;
	case TK_OP_STRING:
	    offs = GET_INT24(p);
	    if (offs >= STR_MAX)
		ErlTk_Exit(1, "bad string offset");
	    argv[i] = tkstr[offs];
	    break;
	default:
	    ErlTk_Exit(1, "bad option type");
	}
    }

    switch(ErlTk_Execute(cmd,(ClientData)mainWindow,interp,argc,argv)) {
    case TCL_BREAK:
	return;

    case TCL_OK:
	code = TK_ERL_OK;
	break;

    case TCL_ERROR:
	code = TK_ERL_ERROR;
	break;
    }
    len =  strlen(interp->result);

    switch(ERL_PACKET_HEADER) {
    case 1:
	PUT_INT8(lenbuf, len+4);
	break;
    case 2:
	PUT_INT16(lenbuf, len+4);
	break;
    case 4:
	PUT_INT32(lenbuf, len+4);
	break;
    }
    erl_write_fill(ERL_OUTPUT, lenbuf, ERL_PACKET_HEADER);

    PUT_INT16(iobuffer, seq);
    PUT_INT16(iobuffer+2, code);
    memcpy(iobuffer+4, interp->result, len);

    Tcl_ResetResult(interp);

    erl_write_fill(ERL_OUTPUT, iobuffer, len+4);
}

/*
** Port main loop
*/
int main(argc, argv)
int argc; char** argv;
{
    /* Initialize */
    int code;
    char* av[20];
    int   ac;
    char* class;
    Colormap cmap;
    Visual* vis;

    erl_closed = 0;
    interp = Tcl_CreateInterp();

    synchronize = 0;
    withdrawn = 0;
    name = display = geometry = colormap = visual = NULL; 

    if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 
		     TK_ARGV_NO_DEFAULTS) != TCL_OK) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }

    OpenLog(logfile);

    if (name == NULL) {
	char* p = strrchr(argv[0], '/');
	if (p == NULL)
	    name = argv[0];
	else
	    name = p+1;
    }

    class = (char *) ckalloc((unsigned) (strlen(name) + 1));
    strcpy(class, name);
    class[0] = toupper((unsigned char) class[0]);

    av[0] = "toplevel";
    av[1] = ".";
    av[2] = "-class";
    av[3] = class;
    ac = 4;
 
    if (display != NULL) {
	av[ac] = "-screen";
	av[ac+1] = display;
	ac += 2;
    }
    if (colormap != NULL) {
	av[ac] = "-colormap";
	av[ac+1] = colormap;
	ac += 2;
    }
    if (visual != NULL) {
	av[ac] = "-visual";
	av[ac+1] = visual;
	ac += 2;
    }
    av[ac] = NULL;
    code = TkCreateFrame((ClientData) NULL, interp, ac, av, 1, name);
    ckfree(class);
    if (code != TCL_OK) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }
    mainWindow = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "event", ErlangEvent, 
		      (ClientData) ERL_OUTPUT, (void (*)()) NULL);

    Tcl_CreateCommand(interp, "invoke", ErlangInvoke, 
		      (ClientData) ERL_OUTPUT, (void (*)()) NULL);

    Tcl_CreateCommand(interp, "bgerror", tkError, 
		      (ClientData) ERL_OUTPUT, (void (*)()) NULL);

    Tcl_CreateCommand(interp, "tkScreenChanged", tkScreenChanged, 
		      (ClientData) ERL_OUTPUT, (void (*)()) NULL);

    Tcl_ResetResult(interp);

    if (synchronize) {
	XSynchronize(Tk_Display(mainWindow), True);
    }

    if (geometry != NULL) {
	av[0] = "wm";
	av[1] = "geometry";
	av[2] = ".";
	av[3] = geometry;
	av[4] = NULL;

	if (Tk_WmCmd((ClientData) mainWindow, interp, 4, av) != TCL_OK) {
	    fprintf(stderr, "%s\n", interp->result);
	    exit(1);
	}
    }

    if (withdrawn) {
	av[0] = "wm";
	av[1] = "withdraw";
	av[2] = ".";
	av[3] = NULL;
	if (Tk_WmCmd((ClientData) mainWindow, interp, 3, av) != TCL_OK) {
	    fprintf(stderr, "%s\n", interp->result);
	    exit(1);
	}
    }


    TkCreateXEventSource(); 

    Tcl_CreateFileHandler(Tcl_GetFile((ClientData)ERL_INPUT, TCL_UNIX_FD),
			  TK_READABLE, ErlangInputProc,
			  (ClientData) ERL_INPUT);


    while (Tk_GetNumMainWindows() > 0) {
        Tcl_DoOneEvent(0);
    }

    ErlTk_Exit(0, "done");
}

